"""Models for Automower Connect API - Planner."""

from dataclasses import dataclass, field
from datetime import datetime
from enum import StrEnum

from mashumaro import DataClassDictMixin, field_options

from .utils import convert_timestamp_to_aware_datetime


class Actions(StrEnum):
    """Actions in the planner of lawn mower."""

    NOT_ACTIVE = "not_active"
    FORCE_PARK = "force_park"
    FORCE_MOW = "force_mow"


@dataclass
class Override(DataClassDictMixin):
    """DataClass for Override values."""

    action: Actions = field(
        metadata=field_options(deserialize=lambda x: Actions(x.lower()))
    )


class RestrictedReasons(StrEnum):
    """Restricted reasons in the planner of lawn mower."""

    NONE = "none"
    WEEK_SCHEDULE = "week_schedule"
    PARK_OVERRIDE = "park_override"
    SENSOR = "sensor"
    DAILY_LIMIT = "daily_limit"
    FOTA = "fota"
    FROST = "frost"
    ALL_WORK_AREAS_COMPLETED = "all_work_areas_completed"
    EXTERNAL = "external"
    NOT_APPLICABLE = "not_applicable"


class ExternalReasons(StrEnum):
    """External reasons for restrictions."""

    GOOGLE_ASSISTANT = "google_assistant"
    AMAZON_ALEXA = "amazon_alexa"
    HOME_ASSISTANT = "home_assistant"
    IFTTT_CALENDAR_CONNECTION = "ifttt_calendar_connection"
    IFTTT = "ifttt"
    SMART_ROUTINE_RAIN_GUARD = "smart_routine_rain_guard"
    SMART_ROUTINE_WILDLIFE_PROTECTION = "smart_routine_wildlife_protection"
    SMART_ROUTINE_FROST_GUARD = "smart_routine_frost_guard"
    SMART_ROUTINE = "smart_routine"
    GARDENA_SMART_SYSTEM = "gardena_smart_system"
    IFTTT_APPLETS = "ifttt_applets"
    DEVELOPER_PORTAL = "developer_portal"


def resolve_external_reason(reason_id: int) -> ExternalReasons | None:  # noqa: C901
    """Resolve the external reason based on the reason ID."""
    if 1000 <= reason_id <= 1999:
        return ExternalReasons.GOOGLE_ASSISTANT
    if 2000 <= reason_id <= 2999:
        return ExternalReasons.AMAZON_ALEXA
    if 3000 <= reason_id <= 3999:
        return ExternalReasons.HOME_ASSISTANT
    if reason_id == 4002:
        return ExternalReasons.IFTTT_CALENDAR_CONNECTION
    if 4000 <= reason_id <= 4999:
        return ExternalReasons.IFTTT
    if 5000 <= reason_id <= 5999:
        return ExternalReasons.GARDENA_SMART_SYSTEM
    if reason_id == 6000:
        return ExternalReasons.SMART_ROUTINE_RAIN_GUARD
    if reason_id == 6001:
        return ExternalReasons.SMART_ROUTINE_FROST_GUARD
    if reason_id == 6500:
        return ExternalReasons.SMART_ROUTINE_WILDLIFE_PROTECTION
    if 6000 <= reason_id <= 6999:
        return ExternalReasons.SMART_ROUTINE
    if 100000 <= reason_id <= 199999:
        return ExternalReasons.IFTTT_APPLETS
    if 200000 <= reason_id <= 299999:
        return ExternalReasons.DEVELOPER_PORTAL
    return None


@dataclass
class Planner(DataClassDictMixin):
    """DataClass for Planner values."""

    next_start_datetime: datetime | None = field(
        metadata=field_options(
            deserialize=convert_timestamp_to_aware_datetime,
            alias="nextStartTimestamp",
        ),
    )
    override: Override
    restricted_reason: RestrictedReasons = field(
        metadata=field_options(
            deserialize=lambda x: RestrictedReasons(x.lower()),
            alias="restrictedReason",
        )
    )
    external_reason: ExternalReasons | None = field(
        metadata=field_options(
            alias="externalReason",
            deserialize=lambda x: resolve_external_reason(int(x)),
        ),
        default=None,
    )



"""A timeline is a set of events on a calendar.

This timeline is used to iterate over program runtime events, managing
the logic for interpreting recurring events for the Husqvarna Automower
calendar. Only upcoming events will be shown.
"""

import datetime
import logging
from collections.abc import Iterable
from dataclasses import dataclass

from dateutil import rrule
from ical.iter import (
    LazySortableItem,
    RecurIterable,
    SortableItem,
    SortableItemTimeline,
)
from ical.timespan import Timespan

from .const import DayOfWeek, ProgramFrequency

__all__ = ["ProgramEvent", "ProgramTimeline"]

_LOGGER = logging.getLogger(__name__)

RRULE_WEEKDAY = {
    DayOfWeek.MONDAY: rrule.MO,
    DayOfWeek.TUESDAY: rrule.TU,
    DayOfWeek.WEDNESDAY: rrule.WE,
    DayOfWeek.THURSDAY: rrule.TH,
    DayOfWeek.FRIDAY: rrule.FR,
    DayOfWeek.SATURDAY: rrule.SA,
    DayOfWeek.SUNDAY: rrule.SU,
}


@dataclass
class ProgramEvent:
    """An instance of a program event."""

    schedule_no: int
    work_area_id: int | None
    start: datetime.datetime
    end: datetime.datetime
    rule: rrule.rrule | None = None

    @property
    def rrule_str(self) -> str | None:
        """Return the recurrence rule string."""
        rule_str = str(self.rule)
        if not self.rule or "DTSTART:" not in rule_str or "RRULE:" not in rule_str:
            return None
        parts = str(self.rule).split("\n")
        if len(parts) != 2:
            return None
        return parts[1].lstrip("RRULE:")  # noqa: B005


class ProgramTimeline(SortableItemTimeline[ProgramEvent]):
    """A timeline of events in an irrigation program."""


def create_recurrence(
    *,
    schedule_no: int,
    work_area_id: int | None,
    frequency: ProgramFrequency,
    dtstart: datetime.datetime,
    duration: datetime.timedelta,
    days_of_week: set[DayOfWeek],
) -> Iterable[SortableItem[Timespan, ProgramEvent]]:
    """Create a timeline using a recurrence rule."""
    byweekday = [RRULE_WEEKDAY[day_of_week] for day_of_week in days_of_week]
    ruleset = rrule.rruleset()
    rule: rrule.rrule
    if frequency == ProgramFrequency.DAILY:
        # Create a RRULE that is FREQ=DAILY
        rule = rrule.rrule(
            freq=rrule.DAILY,
            dtstart=dtstart,
            cache=True,
        )
    elif frequency == ProgramFrequency.WEEKLY:
        # Create a RRULE that is FREQ=WEEKLY
        rule = rrule.rrule(
            freq=rrule.WEEKLY,
            byweekday=byweekday,
            dtstart=dtstart,
            cache=True,
        )
    ruleset.rrule(rule)

    def adapter(
        dtstart: datetime.datetime | datetime.date,
    ) -> SortableItem[Timespan, ProgramEvent]:
        if not isinstance(dtstart, datetime.datetime):
            msg = "Expected datetime, got date"
            raise TypeError(msg)
        dtend = dtstart + duration

        def build() -> ProgramEvent:
            return ProgramEvent(schedule_no, work_area_id, dtstart, dtend, rule)

        return LazySortableItem(Timespan.of(dtstart, dtend), build)

    return RecurIterable(adapter, ruleset)



"""Models for Automower Connect API websocket events."""

from typing import Any, TypedDict

from .model_message import Message


class GenericEventData(TypedDict):
    """Generic websocket event."""

    id: str
    type: str
    attributes: Any


class CuttingHeight(TypedDict):
    """Cutting height information."""

    height: int


class CuttingHeightAttributes(TypedDict):
    """Cutting height attributes."""

    cuttingHeight: CuttingHeight


class MessageAttributes(TypedDict):
    """Position attributes."""

    message: Message


class Position(TypedDict):
    """Position information."""

    latitude: float
    longitude: float


class PositionAttributes(TypedDict):
    """Position attributes."""

    position: Position


class HeadLightAttributes(TypedDict):
    """Head light attributes."""

    headlights: dict[str, str]



"""Models for Automower Connect API - Mower."""

from dataclasses import dataclass, field
from datetime import datetime
from enum import StrEnum
from re import sub

from mashumaro import DataClassDictMixin, field_options

from .utils import convert_timestamp_to_aware_datetime

ERRORCODES = {
    0: "Unexpected error",
    1: "Outside working area",
    2: "No loop signal",
    3: "Wrong loop signal",
    4: "Loop sensor problem, front",
    5: "Loop sensor problem, rear",
    6: "Loop sensor problem, left",
    7: "Loop sensor problem, right",
    8: "Wrong PIN code",
    9: "Trapped",
    10: "Upside down",
    11: "Low battery",
    12: "Empty battery",
    13: "No drive",
    14: "Mower lifted",
    15: "Lifted",
    16: "Stuck in charging station",
    17: "Charging station blocked",
    18: "Collision sensor problem, rear",
    19: "Collision sensor problem, front",
    20: "Wheel motor blocked, right",
    21: "Wheel motor blocked, left",
    22: "Wheel drive problem, right",
    23: "Wheel drive problem, left",
    24: "Cutting system blocked",
    25: "Cutting system blocked",
    26: "Invalid sub-device combination",
    27: "Settings restored",
    28: "Memory circuit problem",
    29: "Slope too steep",
    30: "Charging system problem",
    31: "STOP button problem",
    32: "Tilt sensor problem",
    33: "Mower tilted",
    34: "Cutting stopped - slope too steep",
    35: "Wheel motor overloaded, right",
    36: "Wheel motor overloaded, left",
    37: "Charging current too high",
    38: "Electronic problem",
    39: "Cutting motor problem",
    40: "Limited cutting height range",
    41: "Unexpected cutting height adj",
    42: "Limited cutting height range",
    43: "Cutting height problem, drive",
    44: "Cutting height problem, curr",
    45: "Cutting height problem, dir",
    46: "Cutting height blocked",
    47: "Cutting height problem",
    48: "No response from charger",
    49: "Ultrasonic problem",
    50: "Guide 1 not found",
    51: "Guide 2 not found",
    52: "Guide 3 not found",
    53: "GPS navigation problem",
    54: "Weak GPS signal",
    55: "Difficult finding home",
    56: "Guide calibration accomplished",
    57: "Guide calibration failed",
    58: "Temporary battery problem",
    59: "Temporary battery problem",
    60: "Temporary battery problem",
    61: "Temporary battery problem",
    62: "Battery restriction due to ambient temperature",
    63: "Temporary battery problem",
    64: "Temporary battery problem",
    65: "Temporary battery problem",
    66: "Battery problem",
    67: "Battery problem",
    68: "Temporary battery problem",
    69: "Alarm! Mower switched off",
    70: "Alarm! Mower stopped",
    71: "Alarm! Mower lifted",
    72: "Alarm! Mower tilted",
    73: "Alarm! Mower in motion",
    74: "Alarm! Outside geofence",
    75: "Connection changed",
    76: "Connection NOT changed",
    77: "Com board not available",
    78: "Slipped - Mower has Slipped.Situation not solved with moving pattern",
    79: "Invalid battery combination - Invalid combination of different battery types.",
    80: "Cutting system imbalance    Warning",
    81: "Safety function faulty",
    82: "Wheel motor blocked, rear right",
    83: "Wheel motor blocked, rear left",
    84: "Wheel drive problem, rear right",
    85: "Wheel drive problem, rear left",
    86: "Wheel motor overloaded, rear right",
    87: "Wheel motor overloaded, rear left",
    88: "Angular sensor problem",
    89: "Invalid system configuration",
    90: "No power in charging station",
    91: "Switch cord problem",
    92: "Work area not valid",
    93: "No accurate position from satellites",
    94: "Reference station communication problem",
    95: "Folding sensor activated",
    96: "Right brush motor overloaded",
    97: "Left brush motor overloaded",
    98: "Ultrasonic Sensor 1 defect",
    99: "Ultrasonic Sensor 2 defect",
    100: "Ultrasonic Sensor 3 defect",
    101: "Ultrasonic Sensor 4 defect",
    102: "Cutting drive motor 1 defect",
    103: "Cutting drive motor 2 defect",
    104: "Cutting drive motor 3 defect",
    105: "Lift Sensor defect",
    106: "Collision sensor defect",
    107: "Docking sensor defect",
    108: "Folding cutting deck sensor defect",
    109: "Loop sensor defect",
    110: "Collision sensor error",
    111: "No confirmed position",
    112: "Cutting system major imbalance",
    113: "Complex working area",
    114: "Too high discharge current",
    115: "Too high internal current",
    116: "High charging power loss",
    117: "High internal power loss",
    118: "Charging system problem",
    119: "Zone generator problem",
    120: "Internal voltage error",
    121: "High internal temperature",
    122: "CAN error",
    123: "Destination not reachable",
    701: "Connectivity problem",
    702: "Connectivity settings restored",
    703: "Connectivity problem",
    704: "Connectivity problem",
    705: "Connectivity problem",
    706: "Poor signal quality",
    707: "SIM card requires PIN",
    708: "SIM card locked",
    709: "SIM card not found",
    710: "SIM card locked",
    711: "SIM card locked",
    712: "SIM card locked",
    713: "Geofence problem",
    714: "Geofence problem",
    715: "Connectivity problem",
    716: "Connectivity problem",
    717: "SMS could not be sent",
    724: "Communication circuit board SW must be updated",
}


class InactiveReasons(StrEnum):
    """Inactive reasons why the mower is not working."""

    NONE = "none"
    PLANNING = "planning"
    SEARCHING_FOR_SATELLITES = "searching_for_satellites"


class MowerActivities(StrEnum):
    """Mower activities of a lawn mower."""

    UNKNOWN = "unknown"
    NOT_APPLICABLE = "not_applicable"
    MOWING = "mowing"
    GOING_HOME = "going_home"
    CHARGING = "charging"
    LEAVING = "leaving"
    PARKED_IN_CS = "parked_in_cs"
    STOPPED_IN_GARDEN = "stopped_in_garden"


class MowerModes(StrEnum):
    """Mower activities of a lawn mower."""

    MAIN_AREA = "main_area"
    DEMO = "demo"
    SECONDARY_AREA = "secondary_area"
    HOME = "home"
    UNKNOWN = "unknown"


class MowerStates(StrEnum):
    """Mower states of a lawn mower."""

    FATAL_ERROR = "fatal_error"
    ERROR = "error"
    ERROR_AT_POWER_UP = "error_at_power_up"
    NOT_APPLICABLE = "not_applicable"
    UNKNOWN = "unknown"
    STOPPED = "stopped"
    OFF = "off"
    PAUSED = "paused"
    IN_OPERATION = "in_operation"
    WAIT_UPDATING = "wait_updating"
    WAIT_POWER_UP = "wait_power_up"
    RESTRICTED = "restricted"


@dataclass
class Mower(DataClassDictMixin):
    """Information about the mowers current status."""

    mode: MowerModes = field(
        metadata=field_options(deserialize=lambda x: MowerModes(x.lower()))
    )
    activity: MowerActivities = field(
        metadata=field_options(deserialize=lambda x: MowerActivities(x.lower()))
    )
    state: MowerStates = field(
        metadata=field_options(deserialize=lambda x: MowerStates(x.lower()))
    )
    error_code: int = field(metadata=field_options(alias="errorCode"))
    error_key: str | None = field(
        metadata=field_options(
            deserialize=lambda x: None if x == 0 else snake_case(ERRORCODES.get(x)),
            alias="errorCode",
        )
    )
    error_datetime: datetime | None = field(
        metadata=field_options(
            deserialize=convert_timestamp_to_aware_datetime,
            alias="errorCodeTimestamp",
        ),
    )
    inactive_reason: InactiveReasons = field(
        metadata=field_options(
            alias="inactiveReason", deserialize=lambda x: InactiveReasons(x.lower())
        ),
    )
    is_error_confirmable: bool = field(
        metadata=field_options(alias="isErrorConfirmable"), default=False
    )
    work_area_id: int | None = field(
        metadata=field_options(alias="workAreaId"), default=None
    )
    work_area_name: str | None = field(init=False, default=None)

    def __post_init__(self) -> None:
        """Initialize work_area_name to None for later external setting."""
        self.work_area_name = None


def snake_case(string: str | None) -> str:
    """Convert an error text to snake case."""
    if string is None:
        raise TypeError
    return "_".join(
        sub(
            "([A-Z][a-z][,]+)",
            r" \1",
            sub(
                "([A-Z]+)",
                r" \1",
                string.replace("-", " ")
                .replace(",", "")
                .replace(".", "")
                .replace("!", ""),
            ),
        ).split()
    ).lower()


def error_key_list() -> list[str]:
    """Create a list with all possible error keys."""
    codes = [snake_case(error_text) for error_text in ERRORCODES.values()]
    return sorted(codes)


def error_key_dict() -> dict[str, str]:
    """Create a dictionary with error keys and a human friendly text."""
    codes = {}
    for error_text in ERRORCODES.values():
        codes[snake_case(error_text)] = error_text
    return codes



"""Commands for the Automower."""

import datetime
import logging
import zoneinfo
from dataclasses import dataclass

from .auth import AbstractAuth
from .exceptions import (
    FeatureNotSupportedError,
    WorkAreasDifferentError,
)
from .model import (
    Calendar,
    HeadlightModes,
    MowerAttributes,
    Tasks,
)
from .utils import timedelta_to_minutes

_LOGGER = logging.getLogger(__name__)
FEATURE_NOT_SUPPORTED_MSG = "This mower does not support this command."


@dataclass
class AutomowerEndpoint:
    """Endpoint URLs for the AutomowerConnect API."""

    mowers = "mowers/"
    "List data for all mowers linked to a user."

    actions = "mowers/{mower_id}/actions"
    "Accepts actions to control a mower linked to a user."

    calendar = "mowers/{mower_id}/calendar"
    "Update the calendar on the mower."

    messages = "mowers/{mower_id}/messages"
    "List data for all mowers linked to a user."

    settings = "mowers/{mower_id}/settings"
    "Update the settings on the mower."

    reset_cutting_blade_usage_time = (
        "mowers/{mower_id}/statistics/resetCuttingBladeUsageTime"
    )
    """Reset the cutting blade usage time. Same function that is available in the
    Automower Connect app. The statistics value cuttingBladeUsageTime will be reset.
    Can be used when cutting blades are changed on the Automower to know when its time
    to the blades next time."""

    stay_out_zones = "mowers/{mower_id}/stayOutZones/{stay_out_id}"
    "Enable or disable the stay-out zone."

    work_area_cutting_height = "mowers/{mower_id}/workAreas/{work_area_id}"
    "This will update cutting height on the work area."

    work_area_calendar = "mowers/{mower_id}/workAreas/{work_area_id}/calendar"
    "Update the calendar for a work area on the mower."

    error_confirm = "mowers/{mower_id}/errors/confirm"
    "Confirm mower non-fatal error"


class WorkAreaSettings:
    """Namespace for work area settings commands."""

    def __init__(
        self,
        client: "MowerCommands",
        mower_id: str,
        work_area_id: int,
    ) -> None:
        """Initialize WorkAreaSettings for a mower's work area."""
        self._client = client
        self.mower_id = mower_id
        self.work_area_id = work_area_id

        # Verify capability
        if not client.data[mower_id].capabilities.work_areas:
            msg = FEATURE_NOT_SUPPORTED_MSG
            raise FeatureNotSupportedError(msg)

    async def cutting_height(
        self,
        cutting_height: int,
    ) -> None:
        """Set the cutting height for this work area."""
        payload = {
            "data": {
                "type": "workArea",
                "id": self.work_area_id,
                "attributes": {"cuttingHeight": cutting_height},
            }
        }
        url = AutomowerEndpoint.work_area_cutting_height.format(
            mower_id=self.mower_id,
            work_area_id=self.work_area_id,
        )
        await self._client.auth.patch_json(url, json=payload)

    async def enabled(
        self,
        *,
        enabled: bool,
    ) -> None:
        """Enable or disable this work area."""
        payload = {
            "data": {
                "type": "workArea",
                "id": self.work_area_id,
                "attributes": {"enable": enabled},
            }
        }
        url = AutomowerEndpoint.work_area_cutting_height.format(
            mower_id=self.mower_id,
            work_area_id=self.work_area_id,
        )
        await self._client.auth.patch_json(url, json=payload)


class MowerCommands:
    """Sending commands."""

    def __init__(
        self,
        auth: AbstractAuth,
        data: dict[str, MowerAttributes],
        mower_tz: zoneinfo.ZoneInfo,
    ) -> None:
        """Send all commands to the API.

        :param class auth: The AbstractAuth class from aioautomower.auth.
        """
        self.auth = auth
        self.data = data
        self.mower_tz = mower_tz

    def workarea_settings(
        self,
        mower_id: str,
        work_area_id: int,
    ) -> WorkAreaSettings:
        """Return a settings helper for a specific work area.

        :param mower_id: Identifier of the mower.
        :param work_area_id: Identifier of the work area.
        :returns: Configured WorkAreaSettings instance.
        """
        return WorkAreaSettings(
            client=self,
            mower_id=mower_id,
            work_area_id=work_area_id,
        )

    async def reset_cutting_blade_usage_time(self, mower_id: str) -> None:
        """Reset the cutting blade usage time.

        Same function that is available in the Automower Connect app. The statistics
        value cuttingBladeUsageTime will be reset. Can be used when cutting blades are
        changed on the Automower to know when its time to the blades next time.
        """
        url = AutomowerEndpoint.reset_cutting_blade_usage_time.format(mower_id=mower_id)
        await self.auth.post_json(url)

    async def resume_schedule(self, mower_id: str) -> None:
        """Resume schedule.

        Remove any override on the Planner and let the mower
        resume to the schedule set by the Calendar.
        """
        body = {"data": {"type": "ResumeSchedule"}}
        url = AutomowerEndpoint.actions.format(mower_id=mower_id)
        await self.auth.post_json(url, json=body)

    async def pause_mowing(self, mower_id: str) -> None:
        """Send pause mowing command to the mower via Rest."""
        body = {"data": {"type": "Pause"}}
        url = AutomowerEndpoint.actions.format(mower_id=mower_id)
        await self.auth.post_json(url, json=body)

    async def park_until_next_schedule(self, mower_id: str) -> None:
        """Send park until next schedule command to the mower."""
        body = {"data": {"type": "ParkUntilNextSchedule"}}
        url = AutomowerEndpoint.actions.format(mower_id=mower_id)
        await self.auth.post_json(url, json=body)

    async def park_until_further_notice(self, mower_id: str) -> None:
        """Send park until further notice command to the mower."""
        body = {"data": {"type": "ParkUntilFurtherNotice"}}
        url = AutomowerEndpoint.actions.format(mower_id=mower_id)
        await self.auth.post_json(url, json=body)

    async def park_for(self, mower_id: str, tdelta: datetime.timedelta) -> None:
        """Parks the mower for a period of minutes.

        The mower will drive to
        the charching station and park for the duration set by the command.
        """
        body = {
            "data": {
                "type": "Park",
                "attributes": {"duration": timedelta_to_minutes(tdelta)},
            }
        }
        url = AutomowerEndpoint.actions.format(mower_id=mower_id)
        await self.auth.post_json(url, json=body)

    async def start_in_workarea(
        self,
        mower_id: str,
        work_area_id: int,
        tdelta: datetime.timedelta,
    ) -> None:
        """Start the mower in a work area for a period of minutes."""
        if not self.data[mower_id].capabilities.work_areas:
            msg = FEATURE_NOT_SUPPORTED_MSG
            raise FeatureNotSupportedError(msg)
        body = {
            "data": {
                "type": "StartInWorkArea",
                "attributes": {
                    "duration": timedelta_to_minutes(tdelta),
                    "workAreaId": work_area_id,
                },
            }
        }
        url = AutomowerEndpoint.actions.format(mower_id=mower_id)
        await self.auth.post_json(url, json=body)

    async def start_for(self, mower_id: str, tdelta: datetime.timedelta) -> None:
        """Start the mower for a period of minutes."""
        body = {
            "data": {
                "type": "Start",
                "attributes": {"duration": timedelta_to_minutes(tdelta)},
            }
        }
        url = AutomowerEndpoint.actions.format(mower_id=mower_id)
        await self.auth.post_json(url, json=body)

    async def set_cutting_height(self, mower_id: str, cutting_height: int) -> None:
        """Set the cutting height for the mower."""
        body = {
            "data": {
                "type": "settings",
                "attributes": {"cuttingHeight": cutting_height},
            }
        }
        url = AutomowerEndpoint.settings.format(mower_id=mower_id)
        await self.auth.post_json(url, json=body)

    async def set_datetime(
        self, mower_id: str, current_time: datetime.datetime | None = None
    ) -> None:
        """Set the datetime of the mower.

        Timestamp in seconds from 1970-01-01. The timestamp needs to be in 24 hours in
        the local time of the mower.
        """
        current_time = current_time or datetime.datetime.now(tz=self.mower_tz)
        body = {
            "data": {
                "type": "settings",
                "attributes": {
                    "dateTime": int(
                        current_time.astimezone(self.mower_tz)
                        .replace(tzinfo=datetime.UTC)
                        .timestamp()
                    )
                },
            }
        }
        url = AutomowerEndpoint.settings.format(mower_id=mower_id)
        await self.auth.post_json(url, json=body)

    async def set_datetime_new(
        self, mower_id: str, current_time: datetime.datetime | None = None
    ) -> None:
        """Set the datetime of the mower.

        If the current has not tz_info, the mower_tz will be used as tz_info.
        """
        current_time = current_time or datetime.datetime.now(tz=self.mower_tz)
        body = {
            "data": {
                "type": "settings",
                "attributes": {
                    "timer": {
                        "dateTime": int(
                            current_time.astimezone(self.mower_tz)
                            .replace(tzinfo=datetime.UTC)
                            .timestamp()
                        ),
                        "timeZone": str(self.mower_tz),
                    },
                },
            }
        }
        url = AutomowerEndpoint.settings.format(mower_id=mower_id)
        await self.auth.post_json(url, json=body)

    async def set_headlight_mode(
        self,
        mower_id: str,
        headlight_mode: HeadlightModes,
    ) -> None:
        """Send headlight mode to the mower."""
        if not self.data[mower_id].capabilities.headlights:
            msg = FEATURE_NOT_SUPPORTED_MSG
            raise FeatureNotSupportedError(msg)
        body = {
            "data": {
                "type": "settings",
                "attributes": {"headlight": {"mode": headlight_mode.upper()}},
            }
        }
        url = AutomowerEndpoint.settings.format(mower_id=mower_id)
        await self.auth.post_json(url, json=body)

    async def set_calendar(
        self,
        mower_id: str,
        tasks: Tasks,
    ) -> None:
        """Send calendar task to the mower."""
        if not self.data[mower_id].capabilities.work_areas:
            body = {
                "data": {
                    "type": "calendar",
                    "attributes": tasks.to_dict(),
                }
            }
            url = AutomowerEndpoint.calendar.format(mower_id=mower_id)
            await self.auth.post_json(url, json=body)
        if self.data[mower_id].capabilities.work_areas:
            task_list: list[Calendar] = tasks.tasks
            first_work_area_id = None
            for task in task_list:
                work_area_id = task.work_area_id
                if first_work_area_id is None:
                    first_work_area_id = work_area_id
                elif work_area_id != first_work_area_id:
                    msg = "Only identical work areas are allowed in one command."
                    raise WorkAreasDifferentError(msg)
            body = {
                "data": {
                    "type": "calendar",
                    "attributes": tasks.to_dict(),
                }
            }
            url = AutomowerEndpoint.work_area_calendar.format(
                mower_id=mower_id, work_area_id=work_area_id
            )
            await self.auth.post_json(url, json=body)

    async def switch_stay_out_zone(
        self, mower_id: str, stay_out_zone_id: str, *, switch: bool
    ) -> None:
        """Enable or disable a stay out zone."""
        if not self.data[mower_id].capabilities.stay_out_zones:
            msg = FEATURE_NOT_SUPPORTED_MSG
            raise FeatureNotSupportedError(msg)
        body = {
            "data": {
                "type": "stayOutZone",
                "id": stay_out_zone_id,
                "attributes": {"enable": switch},
            }
        }
        url = AutomowerEndpoint.stay_out_zones.format(
            mower_id=mower_id, stay_out_id=stay_out_zone_id
        )
        await self.auth.patch_json(url, json=body)

    async def error_confirm(self, mower_id: str) -> None:
        """Confirm non-fatal mower error."""
        if not self.data[mower_id].capabilities.can_confirm_error:
            msg = FEATURE_NOT_SUPPORTED_MSG
            raise FeatureNotSupportedError(msg)
        url = AutomowerEndpoint.error_confirm.format(mower_id=mower_id)
        await self.auth.post_json(url)



"""Models for Automower Connect API - Positions."""

from dataclasses import dataclass

from mashumaro import DataClassDictMixin


@dataclass
class Positions(DataClassDictMixin):
    """List of the GPS positions.

    Latest registered position is first in the
    array and the oldest last in the array.
    Max number of positions is 50 after
    that the latest position is removed
    from the array.
    """

    latitude: float
    longitude: float



"""Provide a model for the Automower Connect API."""

from .model import (
    MowerAttributes,
    MowerData,
    MowerDictionary,
    MowerList,
    generate_work_area_dict,
    generate_work_area_names_list,
    get_work_area_name,
)
from .model_battery import Battery
from .model_calendar import (
    WEEKDAYS,
    WEEKDAYS_TO_ICAL,
    AutomowerCalendarEvent,
    Calendar,
    ConvertScheduleToCalendar,
    DurationSerializationStrategy,
    Tasks,
    TimeSerializationStrategy,
    make_name_string,
)
from .model_capabilities import Capabilities
from .model_message import Message, MessageData, Severity, SingleMessageData
from .model_metadata import Metadata
from .model_mower import (
    InactiveReasons,
    Mower,
    MowerActivities,
    MowerModes,
    MowerStates,
    error_key_dict,
    error_key_list,
)
from .model_planner import (
    Actions,
    ExternalReasons,
    Override,
    Planner,
    RestrictedReasons,
)
from .model_positions import (
    Positions,
)
from .model_settings import (
    Headlight,
    HeadlightModes,
)
from .model_stay_out_zones import StayOutZones, Zone
from .model_system import System
from .model_token import (
    JWT,
    User,
)
from .model_work_areas import WorkArea

__all__ = [
    "JWT",
    "WEEKDAYS",
    "WEEKDAYS_TO_ICAL",
    "Actions",
    "AutomowerCalendarEvent",
    "Battery",
    "Calendar",
    "Capabilities",
    "ConvertScheduleToCalendar",
    "DurationSerializationStrategy",
    "ExternalReasons",
    "Headlight",
    "HeadlightModes",
    "InactiveReasons",
    "Message",
    "MessageData",
    "Metadata",
    "Mower",
    "MowerActivities",
    "MowerAttributes",
    "MowerData",
    "MowerDictionary",
    "MowerList",
    "MowerModes",
    "MowerStates",
    "Override",
    "Planner",
    "Positions",
    "RestrictedReasons",
    "Settings",
    "Severity",
    "SingleMessageData",
    "Statistics",
    "StayOutZones",
    "System",
    "Tasks",
    "TimeSerializationStrategy",
    "User",
    "WorkArea",
    "Zone",
    "convert_timestamp_to_aware_datetime",
    "error_key_dict",
    "error_key_list",
    "generate_work_area_dict",
    "generate_work_area_names_list",
    "get_work_area_name",
    "make_name_string",
]



"""Models for Automower Connect API messages."""

from typing import Literal, TypedDict


class Message(TypedDict):
    """A single message containing error or warning information."""

    time: int
    code: int
    severity: Literal["FATAL", "ERROR", "WARNING", "INFO", "DEBUG", "SW", "UNKNOWN"]
    latitude: float
    longitude: float


class MesssageAttributes(TypedDict):
    """Container for a list of diagnostic messages."""

    messages: list[Message]


class Data(TypedDict):
    """Root data object containing type, ID, and attributes."""

    type: Literal["messages"]
    id: Literal["messages"]
    attributes: MesssageAttributes | None


class MessageResponse(TypedDict):
    """Top-level response structure for messages."""

    data: Data



"""Models for Automower Connect API - Battery."""

from dataclasses import dataclass, field

from mashumaro import DataClassDictMixin, field_options


@dataclass
class Battery(DataClassDictMixin):
    """Information about the battery in the Automower."""

    battery_percent: int = field(metadata=field_options(alias="batteryPercent"))



"""Models for Automower Connect API - WorkAreas."""

from dataclasses import dataclass, field
from datetime import datetime

from mashumaro import DataClassDictMixin, field_options

from .utils import convert_timestamp_to_aware_datetime


@dataclass
class WorkArea(DataClassDictMixin):
    """DataClass for WorkArea values."""

    name: str = field(
        metadata=field_options(
            deserialize=lambda x: "my_lawn" if x == "" else x,
        ),
    )
    cutting_height: int = field(metadata=field_options(alias="cuttingHeight"))
    enabled: bool = field(default=False)
    progress: int | None = field(default=None)
    last_time_completed: datetime | None = field(
        metadata=field_options(
            deserialize=convert_timestamp_to_aware_datetime,
            alias="lastTimeCompleted",
        ),
        default=None,
    )



"""Models for Husqvarna Authentication API."""

from dataclasses import dataclass

from mashumaro import DataClassDictMixin


@dataclass
class User(DataClassDictMixin):
    """The user details of the JWT."""

    first_name: str
    last_name: str
    custom_attributes: dict[str, str]
    customer_id: str


@dataclass
class JWT(DataClassDictMixin):
    """The content of the JWT."""

    jti: str
    iss: str
    roles: list[str]
    groups: list[str]
    scopes: list[str]
    scope: str
    client_id: str
    customer_id: str
    user: User
    iat: int
    exp: int
    sub: str



"""Helper utils for Automower Connect API."""

from datetime import UTC, datetime

from aioautomower import tz_util


def convert_timestamp_to_aware_datetime(timestamp: int) -> datetime | None:
    """Convert the timestamp to an aware datetime object.

    The Python datetime library expects timestamps to be anchored in UTC,
    however, the automower timestamps are anchored in local time. So we convert
    the timestamp to a datetime and replace the timezone with the local time.
    After that we convert the timezone to UTC.
    """
    if timestamp == 0:
        return None
    if timestamp > 32503680000:
        # This will break on January 1th 3000. If mankind still exists there
        # please fix it.
        return datetime.fromtimestamp(timestamp / 1000, tz=UTC).replace(
            tzinfo=tz_util.MOWER_TIME_ZONE
        )
    return datetime.fromtimestamp(timestamp, tz=UTC).replace(
        tzinfo=tz_util.MOWER_TIME_ZONE
    )
